
# -*- coding: utf-8 -*-
from multiprocessing import Process, Manager
import time
import os
from aiohttp import web
import logging
from unittest.mock import MagicMock, patch
import asyncio
import random
from cbpi.api import *
import bluetooth._bluetooth as bluez
from . import blescan
#from bleak import BleakScanner

logger = logging.getLogger(__name__)

import numpy as np

global tilt_cache

tilt_proc = None
tilt_manager = None
tilt_cache = {}
cache = {}

TILTS = {
	'a495bb10c5b14b44b5121370f02d74de': 'Red',
	'a495bb20c5b14b44b5121370f02d74de': 'Green',
	'a495bb30c5b14b44b5121370f02d74de': 'Black',
	'a495bb40c5b14b44b5121370f02d74de': 'Purple',
	'a495bb50c5b14b44b5121370f02d74de': 'Orange',
	'a495bb60c5b14b44b5121370f02d74de': 'Blue',
	'a495bb70c5b14b44b5121370f02d74de': 'Yellow',
	'a495bb80c5b14b44b5121370f02d74de': 'Pink',
}

def add_calibration_point(x, y, field):
    if isinstance(field, str) and field:
        x1, y1 = field.split("=")
        x = np.append(x, float(x1))
        y = np.append(y, float(y1))
    return x, y

def calcGravity(gravity, unitsGravity):
    sg = float(gravity)/1000
    #logging.info(sg)
    #logging.info(unitsGravity)
    if (unitsGravity == "Plato"):
        # Source: https://en.wikipedia.org/wiki/Brix
        return ((135.997 * sg - 630.272) * sg + 1111.14) * sg - 616.868
    elif (unitsGravity == "Brix"):
        # Source: https://en.wikipedia.org/wiki/Brix
        return ((182.4601 * sg - 775.6821) * sg + 1262.7794) * sg - 669.5622
    else:
        return sg

def calcTemp(temp,unit):
	f = float(temp)
	if unit == "C":
		return (f - 32) / 1.8
	else:
		return f

def calibrate(tilt, equation):
    return eval(equation)
	
def distinct(objects):
    seen = set()
    unique = []
    for obj in objects:
        if obj['uuid'] not in seen:
            unique.append(obj)
            seen.add(obj['uuid'])
    return unique

def readTilt(cache):
    dev_id = 0
    while True:
        try:
            logging.info("Starting Bluetooth connection")
            sock = bluez.hci_open_dev(dev_id)
            blescan.hci_le_set_scan_parameters(sock)
            blescan.hci_enable_le_scan(sock)

            while True:
                beacons = distinct(blescan.parse_events(sock, 10))
                #print(beacons)
                for beacon in beacons:
                    if beacon['uuid'] in TILTS.keys():
                        cache[TILTS[beacon['uuid']]] = {'Temp': beacon['major'], 'Gravity': beacon['minor'], 'Time': time.time(),'RSSI': beacon['rssi']}
                        #logging.info(cache)
                        logging.info("Tilt data received: Temp: %s Gravity: %s RSSI: %s" % (beacon['major'], beacon['minor'], beacon['rssi']))
                        time.sleep(4)
        except Exception as e:
            logging.error("Error starting Bluetooth device, exception: %s" % str(e))

        logging.info("Restarting Bluetooth process in 10 seconds")
        time.sleep(10)



@parameters([Property.Select(label="Sensor color", options=["Red", "Green", "Black", "Purple", "Orange", "Blue", "Yellow", "Pink"], description="Select the color of your Tilt"),
	         Property.Select(label= "Data Type", options=["Temperature", "Gravity","RSSI"], description="Select which type of data to register for this sensor"),
	         Property.Select(label="Gravity Units", options=["SG", "Brix", "Plato"], description="Converts the gravity reading to this unit if the Data Type is set to Gravity"),
	         Property.Text(label="Calibration Point 1", configurable=True, default_value="", description="Optional field for calibrating your Tilt. Enter data in the format uncalibrated=actual"),
             Property.Text(label="Calibration Point 2", configurable=True, default_value="", description="Optional field for calibrating your Tilt. Enter data in the format uncalibrated=actual"),
             Property.Text(label="Calibration Point 3", configurable=True, default_value="", description="Optional field for calibrating your Tilt. Enter data in the format uncalibrated=actual")])
class BLESensor(CBPiSensor):
    
    def __init__(self, cbpi, id, props):
        super(BLESensor, self).__init__(cbpi, id, props)
        global titl_cache
        self.value = 0
        self.calibration_equ=""
        self.x_cal_1=self.props.get("Calibration Point 1","")
        self.x_cal_2=self.props.get("Calibration Point 2","")
        self.x_cal_3=self.props.get("Calibration Point 3","")

        self.color=self.props.get("Sensor color","")
        self.sensorType=self.props.get("Data Type","Temperature")
        self.unitsGravity=self.props.get("Gravity Units","Plato")
        self.time_old = float(0)
        # Load calibration data from plugin
        x = np.empty([0])
        y = np.empty([0])
        x, y = add_calibration_point(x, y, self.x_cal_1)
        x, y = add_calibration_point(x, y, self.x_cal_2)
        x, y = add_calibration_point(x, y, self.x_cal_3)

        # Create calibration equation
        if len(x) < 1:
            self.calibration_equ = "tilt"
        if len(x) == 1:
            self.calibration_equ = 'tilt + {0}'.format(y[0] - x[0])
        if len(x) > 1:
            A = np.vstack([x, np.ones(len(x))]).T
            m, c = np.linalg.lstsq(A, y, rcond=None)[0]
            self.calibration_equ = '{0}*tilt + {1}'.format(m, c)

        #logging.info('Calibration equation: {0}'.format(self.calibration_equ))

    async def run(self):
        while self.running is True:
            if self.color in tilt_cache:
                current_time = float(tilt_cache[self.color]['Time'])
                #logging.info("Color: {} | Time Old: {} | Curent Time: {}".format(self.color,self.time_old,current_time))
                if self.sensorType == "Gravity":
                    if current_time > self.time_old:
                        reading = calcGravity(tilt_cache[self.color]['Gravity'], self.unitsGravity)
                        reading = calibrate(reading, self.calibration_equ)
                        reading = round(reading, 3)
                        self.time_old = current_time
                        self.value = reading
                        self.log_data(self.value)
                elif self.sensorType == "Temperature":
                    self.TEMP_UNIT=self.get_config_value("TEMP_UNIT", "C")
                    if current_time > self.time_old:
                        reading = calcTemp(tilt_cache[self.color]['Temp'],self.TEMP_UNIT)
                        reading = round(reading, 2)
                        self.time_old = current_time 
                        self.value=reading
                        self.log_data(self.value)
                        self.push_update(self.value)
                else:
                    if current_time > self.time_old:
                        reading = tilt_cache[self.color]['RSSI']
                        self.time_old = current_time 
                        self.value=reading
                        self.log_data(self.value)
                        self.push_update(self.value)
                        
                self.push_update(self.value,False)

            await asyncio.sleep(2)
    
    def get_state(self):
        return dict(value=self.value)

def setup(cbpi):
    global tilt_proc
    global tilt_manager
    global tilt_cache
    print ("INITIALIZE TILT MODULE")
    
    tilt_manager = Manager()
    tilt_cache = tilt_manager.dict()

    tilt_proc = Process(name='readTilt', target=readTilt, args=(tilt_cache,))
    tilt_proc.daemon = True
    tilt_proc.start()
    cbpi.plugin.register("BLE Hydrom", BLESensor)
    pass



from __future__ import print_function
# BLE iBeaconScanner based on https://github.com/adamf/BLE/blob/master/ble-scanner.py
# SwitchDoc Lab December 2020

from builtins import range
DEBUG = False 
# BLE scanner based on https://github.com/adamf/BLE/blob/master/ble-scanner.py
# BLE scanner, based on https://code.google.com/p/pybluez/source/browse/trunk/examples/advanced/inquiry-with-rssi.py

# https://github.com/pauloborges/bluez/blob/master/tools/hcitool.c for lescan
# https://kernel.googlesource.com/pub/scm/bluetooth/bluez/+/5.6/lib/hci.h for opcodes
# https://github.com/pauloborges/bluez/blob/master/lib/hci.c#L2782 for functions used by lescan

# performs a simple device inquiry, and returns a list of ble advertizements 
# discovered device

# NOTE: Python's struct.pack() will add padding bytes unless you make the endianness explicit. Little endian
# should be used for BLE. Always start a struct.pack() format string with "<"

import os
import sys
import struct
import bluetooth._bluetooth as bluez

LE_META_EVENT = 0x3e
LE_PUBLIC_ADDRESS=0x00
LE_RANDOM_ADDRESS=0x01
LE_SET_SCAN_PARAMETERS_CP_SIZE=7
OGF_LE_CTL=0x08
OCF_LE_SET_SCAN_PARAMETERS=0x000B
OCF_LE_SET_SCAN_ENABLE=0x000C
OCF_LE_CREATE_CONN=0x000D

LE_ROLE_MASTER = 0x00
LE_ROLE_SLAVE = 0x01

# these are actually subevents of LE_META_EVENT
EVT_LE_CONN_COMPLETE=0x01
EVT_LE_ADVERTISING_REPORT=0x02
EVT_LE_CONN_UPDATE_COMPLETE=0x03
EVT_LE_READ_REMOTE_USED_FEATURES_COMPLETE=0x04

# Advertisment event types
ADV_IND=0x00
ADV_DIRECT_IND=0x01
ADV_SCAN_IND=0x02
ADV_NONCONN_IND=0x03
ADV_SCAN_RSP=0x04


def returnnumberpacket(pkt):
    myInteger = 0
    multiple = 256
    for c in pkt:
        myInteger +=  struct.unpack("B",c.to_bytes(1,'little'))[0] * multiple
        multiple = 1
    return myInteger 

def returnstringpacket(pkt):
    myString = "";
    for c in pkt:
        myString +=  "%02x" %struct.unpack("B",c.to_bytes(1,'little'))[0]
    return myString 

def printpacket(pkt):
    for c in pkt:
        sys.stdout.write("%02x " % struct.unpack("B",c.to_bytes(1,'little'))[0])

def get_packed_bdaddr(bdaddr_string):
    packable_addr = []
    addr = bdaddr_string.split(':')
    addr.reverse()
    for b in addr: 
        packable_addr.append(int(b, 16))
    return struct.pack("<BBBBBB", *packable_addr)

def packed_bdaddr_to_string(bdaddr_packed):
    if (len(bdaddr_packed) == 6):
        return ':'.join('%02x'%i for i in struct.unpack("<BBBBBB", bdaddr_packed[::-1]))
    return ':'.join('%02x'%i for i in struct.unpack("<BBBBBBBB", bdaddr_packed[::-1]))

def hci_enable_le_scan(sock):
    hci_toggle_le_scan(sock, 0x01)

def hci_disable_le_scan(sock):
    hci_toggle_le_scan(sock, 0x00)

def hci_toggle_le_scan(sock, enable):
# hci_le_set_scan_enable(dd, 0x01, filter_dup, 1000);
# memset(&scan_cp, 0, sizeof(scan_cp));
 #uint8_t         enable;
 #       uint8_t         filter_dup;
#        scan_cp.enable = enable;
#        scan_cp.filter_dup = filter_dup;
#
#        memset(&rq, 0, sizeof(rq));
#        rq.ogf = OGF_LE_CTL;
#        rq.ocf = OCF_LE_SET_SCAN_ENABLE;
#        rq.cparam = &scan_cp;
#        rq.clen = LE_SET_SCAN_ENABLE_CP_SIZE;
#        rq.rparam = &status;
#        rq.rlen = 1;

#        if (hci_send_req(dd, &rq, to) < 0)
#                return -1;
    cmd_pkt = struct.pack("<BB", enable, 0x00)
    bluez.hci_send_cmd(sock, OGF_LE_CTL, OCF_LE_SET_SCAN_ENABLE, cmd_pkt)


def hci_le_set_scan_parameters(sock):
    old_filter = sock.getsockopt( bluez.SOL_HCI, bluez.HCI_FILTER, 14)

    SCAN_RANDOM = 0x01
    OWN_TYPE = SCAN_RANDOM
    SCAN_TYPE = 0x01


    
def parse_events(sock, loop_count=100):
    old_filter = sock.getsockopt( bluez.SOL_HCI, bluez.HCI_FILTER, 14)

    # perform a device inquiry on bluetooth device #0
    # The inquiry should last 8 * 1.28 = 10.24 seconds
    # before the inquiry is performed, bluez should flush its cache of
    # previously discovered devices
    flt = bluez.hci_filter_new()
    bluez.hci_filter_all_events(flt)
    bluez.hci_filter_set_ptype(flt, bluez.HCI_EVENT_PKT)
    sock.setsockopt( bluez.SOL_HCI, bluez.HCI_FILTER, flt )
    done = False
    results = []
    myFullList = []
    beacons=[]
    for i in range(0, loop_count):
        pkt = sock.recv(255)
        ptype, event, plen = struct.unpack("BBB", pkt[:3])
        #print "--------------" 
        if event == bluez.EVT_INQUIRY_RESULT_WITH_RSSI:
             i =0
        elif event == bluez.EVT_NUM_COMP_PKTS:
             i =0 
        elif event == bluez.EVT_DISCONN_COMPLETE:
             i =0 
        elif event == LE_META_EVENT:

            subevent, = struct.unpack("B", pkt[3].to_bytes(1, 'little'))

            pkt = pkt[4:]
            if subevent == EVT_LE_CONN_COMPLETE:
                le_handle_connection_complete(pkt)
            elif subevent == EVT_LE_ADVERTISING_REPORT:
                #print "advertising report"
                num_reports = struct.unpack("B", pkt[0].to_bytes(1,'little'))[0]
                report_pkt_offset = 0
                for i in range(0, num_reports):
		
                    if (DEBUG == True):
                        print("-------------")
                        #print "\tfullpacket: ", printpacket(pkt)
                        print("\tUUID: ", printpacket(pkt[report_pkt_offset -22: report_pkt_offset - 6]))
                        print("\tMAJOR: ", printpacket(pkt[report_pkt_offset -6: report_pkt_offset - 4]))
                        print("\tMINOR: ", printpacket(pkt[report_pkt_offset -4: report_pkt_offset - 2]))
                        print("\tMAC address: ", packed_bdaddr_to_string(pkt[report_pkt_offset + 3:report_pkt_offset + 9]))
                        # commented out - don't know what this byte is.  It's NOT TXPower
                        txpower, = struct.unpack("b", pkt[report_pkt_offset -2].to_bytes(1,'little'))
                        print("\t(Unknown):", txpower)
	
                        rssi, = struct.unpack("b", pkt[report_pkt_offset -1].to_bytes(1,'little'))
                        print("\tRSSI:", rssi)
                    # build the return string
                    Adstring = packed_bdaddr_to_string(pkt[report_pkt_offset +1:report_pkt_offset + 9])
                    Adstring += ","
                    Adstring += returnstringpacket(pkt[report_pkt_offset -22: report_pkt_offset - 6]) 
                    Adstring += ","
                    Adstring += "%i" % returnnumberpacket(pkt[report_pkt_offset -6: report_pkt_offset - 4]) 
                    Adstring += ","
                    Adstring += "%i" % returnnumberpacket(pkt[report_pkt_offset -4: report_pkt_offset - 2]) 
                    Adstring += ","
                    Adstring += "%i" % struct.unpack("b", pkt[report_pkt_offset -2].to_bytes(1,'little'))
                    Adstring += ","
                    Adstring += "%i" % struct.unpack("b", pkt[report_pkt_offset -1].to_bytes(1,'little'))

                    #print ("\tAdstring={}".format(Adstring))
                    myFullList.append(Adstring)
                    #print(myFullList)
                    beacons.append({
                        'uuid': returnstringpacket(pkt[report_pkt_offset - 22: report_pkt_offset - 6]),
                        'minor': returnnumberpacket(pkt[report_pkt_offset - 4: report_pkt_offset - 2]),
                        'major': returnnumberpacket(pkt[report_pkt_offset - 6: report_pkt_offset - 4]),
                        'rssi': int.from_bytes(pkt[report_pkt_offset -1].to_bytes(1,'little'),byteorder='big', signed=True)
                    })
                done = True
                #print(beacons)
                sock.setsockopt( bluez.SOL_HCI, bluez.HCI_FILTER, old_filter )
                return beacons





from setuptools import setup
import os
from setuptools.command.install import install
from subprocess import check_call


class PreInstallCommand(install):
    """Pre-installation for installation mode."""
    def run(self):
        check_call("apt-get install pkg-config libboost-python-dev libboost-thread-dev libbluetooth-dev libglib2.0-dev python3-dev".split())
        install.run(self)

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(name='cbpi4-BLEHydrom',
      version='0.0.3',
      description='CraftBeerPi4 Plugin for Hydrom and Tilt (BLE connection)',
      author='Alexander Vollkopf',
      author_email='avollkopf@web.de',
      url='',
      include_package_data=True,
      package_data={
        # If any package contains *.txt or *.rst files, include them:
      '': ['*.txt', '*.rst', '*.yaml'],
      'cbpi4-BLEHydrom': ['*','*.txt', '*.rst', '*.yaml']},
      packages=['cbpi4-BLEHydrom'],
      install_requires=[
      'PyBluez==0.23',
      'gattlib==0.20201113',
      ],
      long_description=long_description,
      long_description_content_type='text/markdown'
     )
